/**********************************************
 *          the Beos Reversing Engine         * 
 *                  by Zadig                  *
 *                                            *
 * This file contains the chain-lists         *
 * manipulation functions.                    *
 **********************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "chainlist.h"
#include "misc.h"

/**************************************************************
 *                     Mapping functions                      *
 **************************************************************/ 

/***********************************************************
 *  Function: LIST_GetLowerStringMap    File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Frees all elements of a list.                           *
 ***********************************************************/ 
int32 LIST_GetLowerStringMap(LIST_Mapping *list, Mapping **result)
{
   Mapping *p_Map;

   *result = NULL;
   p_Map = list->list_Map;

   if(p_Map == NULL)
      return(CHAIN_ERROR_NOT_FOUND);

   while(p_Map != NULL) 
   {
      if(p_Map->isPrinted == Z_FALSE)
      {
         if( (*result == NULL) ||
               ( (*result != NULL) && 
                 (strcasecmp(p_Map->label,(*result)->label) < 0) 
               ) )
         {
            *result = p_Map;
         }
      }
      p_Map = p_Map->next;
   }

   if(*result == NULL)
      return(CHAIN_ERROR_NOT_FOUND);

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_GetLowerClassMap    File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Frees all elements of a list.                           *
 ***********************************************************/ 
int32 LIST_GetLowerClassMap(LIST_Mapping *list, Mapping **result, uint8 *sz_ClassName)
{
   uint8 *psz_ClassName;
   uint8 sz_CurClassName[STRING_MAX_SIZE];
   uint16 i_Index;
   Mapping *p_Map;

   *result = NULL;
   p_Map = list->list_Map;

   if(p_Map == NULL)
      return(CHAIN_ERROR_NOT_FOUND);

   while(p_Map != NULL)
   {
      /* looks for new class */
      /* don't process templates */
      psz_ClassName = strstr(p_Map->label,"<"); 
      if( (psz_ClassName == NULL) && /* templates contain '<' and '>' */
          (p_Map->isPrinted == Z_FALSE) &&
          (p_Map->type == OBJ_TYPE_FUNCTION) ) /* don't process if already printed */
      {  
         psz_ClassName = strstr(p_Map->label,"::");
         if( psz_ClassName != NULL)
         {
            i_Index = 0;
            while( (psz_ClassName != p_Map->label) && (*psz_ClassName != ' ') )
            {
               i_Index++;
               psz_ClassName--;
            }
            /* if there is a space in the name, it's not a class */
            if(*psz_ClassName != ' ')
            {
               strncpy(sz_CurClassName,psz_ClassName,i_Index);
               sz_CurClassName[i_Index] = '\0';
               if( (*result == NULL) ||
                     ( (*result != NULL) && 
                       (strcasecmp(sz_CurClassName,sz_ClassName) < 0)
                     ))
               {
                  *result = p_Map;
                  strcpy(sz_ClassName, sz_CurClassName);
               }
            }
         } 
      }    
      p_Map = p_Map->next;
   }

   if(*result == NULL)
      return(CHAIN_ERROR_NOT_FOUND);

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_GetLowerMethodMap    File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Frees all elements of a list.                           *
 ***********************************************************/ 
int32 LIST_GetLowerMethodMap(LIST_Mapping *list, Mapping **result, uint8 *sz_ClassName)
{
   uint8 *psz_MethodName;
   uint16 i_ClassNameSize;
   Mapping *p_Map;

   *result = NULL;
   i_ClassNameSize = strlen(sz_ClassName);
   p_Map = list->list_Map;

   if(p_Map == NULL)
      return(CHAIN_ERROR_NOT_FOUND);

   while(p_Map != NULL)
   {
      if( (p_Map->isPrinted == Z_FALSE) &&
            (p_Map->type == OBJ_TYPE_FUNCTION) &&
            ( (psz_MethodName = strstr(p_Map->label,"::")) != NULL) &&
            (strstr(p_Map->label,"<") == NULL) )
      {
         if( (psz_MethodName - p_Map->label) == i_ClassNameSize)
         {
            if( strncmp(p_Map->label,sz_ClassName, i_ClassNameSize) == 0) 
            {
               if( (*result == NULL) ||
                     ( (*result != NULL) && 
                       (strcasecmp(p_Map->label,(*result)->label) < 0) 
                     ) )
               {
                  *result = p_Map;
               }
            }
         } /* end if( (psz_MethodName - p_Map->label) == i_Index) */
      } /* end if( (p_Map->isPrinted == Z_FALSE) &&...*/ 
      p_Map = p_Map->next;
   }

   if(*result == NULL)
      return(CHAIN_ERROR_NOT_FOUND);

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_FreeMap              File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Frees all elements of a list.                           *
 ***********************************************************/ 
int32 LIST_FreeMap(LIST_Mapping *list)
{
   Mapping *p_Next,*p_Map = list->list_Map;

   if( p_Map == NULL )
   {
      return(Z_OK);
   }

   while (p_Map->next != NULL)
   {
      p_Next = p_Map->next;
      LIST_FreeRef(&(p_Map->references));
      free(p_Map);
      p_Map=p_Next;

   }

   free(p_Map);
   list->list_Map = NULL;

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_DeleteMap            File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Removes an element of the list.                         *
 ***********************************************************/ 
int32 LIST_DeleteMap(LIST_Mapping *list, Mapping *mapping)
{
   /*int32 i;
     Mapping *p_Map, *p_Prev = NULL;*/

   mapping->prev->next = mapping->next;
   mapping->next->prev = mapping->prev;

   if(list->list_Map == mapping)    
      list->list_Map = mapping->next;

   if(list->tail == mapping)
      list->tail = mapping->prev;

   free(mapping);

   /*i = 0;
     p_Map = list->list_Map;
     while (i < i_Index)
     {
     p_Prev = p_Map;
     p_Map = p_Map->next;
     i++;
     }

     if (p_Prev != NULL)
     {
     if(p_Map != NULL)
     {
     p_Prev->next = p_Map->next;
     p_Map->next->prev = p_Prev;
     }
     else
     p_Prev->next = NULL;
     }
     else
     {
     list->list_Map = p_Map->next;
     p_Map->prev = NULL;
     }

     free(p_Map);*/

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_AddMap            File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Adds an element in the list and sort it by address.     *
 ***********************************************************/ 
int32 LIST_AddMap(LIST_Mapping *list, Mapping map)
{
   Mapping *p_Mess, *p_Prev, *p_Cur;

   p_Prev = NULL;
   /* Allocates the object */
   p_Mess = (Mapping*)malloc(sizeof(Mapping));

   if (list->list_Map == NULL)
   {
      list->list_Map = p_Mess; /*(Mapping*) malloc(sizeof(Mapping));*/
      list->list_Map->addr = map.addr;
      strcpy(list->list_Map->label, map.label);	
      list->list_Map->type = map.type;
      memcpy(&(list->list_Map->references),&(map.references),sizeof(LIST_Reference));
      list->list_Map->isPrinted = Z_FALSE;
      list->list_Map->next = NULL;
      list->list_Map->prev = NULL;
      list->list_Map->references.list_Ref = NULL; 		
      list->tail = list->list_Map;
      return(Z_OK);
   }

   p_Cur = list->tail;
   if (map.addr >= p_Cur->addr)
   {
      p_Mess->next = NULL;
      p_Mess->prev = p_Cur;
      p_Cur->next = p_Mess;
      list->tail = p_Mess;
   }
   else
   {
      while( (map.addr < p_Cur->addr) &&
            (p_Cur->prev != NULL) )
      {
         p_Prev = p_Cur;
         p_Cur = p_Cur->prev;
      }

      if( (p_Cur->prev == NULL)	&&
            (map.addr < p_Cur->addr) )
      {
         p_Cur->prev = p_Mess;
         p_Mess->next = p_Cur;
         p_Mess->prev = NULL;
         list->list_Map = p_Mess;
      }
      else
      {
         p_Mess->next = p_Prev;
         p_Mess->prev = p_Cur;
         p_Prev->prev = p_Mess;
         p_Cur->next = p_Mess;
      }
   }

   p_Mess->addr = map.addr;
   strcpy(p_Mess->label, map.label);	
   p_Mess->type = map.type;
   p_Mess->references.list_Ref = NULL;
   p_Mess->isPrinted = Z_FALSE;

   return(Z_OK);;

}

/***********************************************************
 *  Function: LIST_FreeMap              File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Gets the element of the list which is at address i_Addr.*
 ***********************************************************/ 
int32 LIST_GetMappingAtAddr(LIST_Mapping *list, uint32 i_Addr, Mapping **map)
{
   uint8 b_Found = Z_FALSE;
   Mapping *p_CurMap;

   p_CurMap = list->list_Map;
   *map = NULL;

   while( (p_CurMap != NULL) && (b_Found == Z_FALSE) )
   {
      if(p_CurMap->addr == i_Addr)
      {
         b_Found = Z_TRUE;
         *map = p_CurMap;
      }
      p_CurMap = p_CurMap->next;
   }

   return(Z_OK);
}


/**************************************************************
 *                    Reference functions                     *
 **************************************************************/ 

/***********************************************************
 *  Function: LIST_FreeRef              File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Frees all elements of a list.                           *
 ***********************************************************/ 
int32 LIST_FreeRef(LIST_Reference *list)
{
   Reference *p_Next,*p_Ref = list->list_Ref;

   if( p_Ref == NULL )
   {
      return(Z_OK);
   }

   while (p_Ref->next != NULL)
   {
      p_Next = p_Ref->next;
      free(p_Ref);
      p_Ref=p_Next;

   }

   free(p_Ref);
   list->list_Ref = NULL;

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_DeleteRef            File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Removes an element of the list.                         *
 ***********************************************************/ 
int32 LIST_DeleteRef(LIST_Reference *list, int32 i_Index)
{
   int32 i=0;
   Reference *p_Ref, *p_Prev;

   if (i_Index == 0) 
   {
      p_Prev = NULL;
      p_Ref = list->list_Ref;
   }
   else 
   {
      p_Ref = list->list_Ref;
      while (i++ < i_Index-1)
      {
         p_Ref = p_Ref->next;
      }
      p_Prev = p_Ref;
      p_Ref = p_Ref->next;
   }

   if (p_Prev)
   {
      p_Prev->next = p_Ref->next;
   }
   else
   {
      list->list_Ref = p_Ref->next;
   }

   free(p_Ref);

   return(Z_OK);
}

/***********************************************************
 *  Function: LIST_AppendRef            File: chainlist.c  *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Appends an element to the list.                         *
 ***********************************************************/ 
int32 LIST_AppendRef(LIST_Reference *list, Reference ref)
{
   Reference *p_List,*p_Mess= list->list_Ref;

   if (list->list_Ref == NULL)
   {
      list->list_Ref = (Reference*) malloc(sizeof(Reference));
      list->list_Ref->addr = ref.addr;
      list->list_Ref->type = ref.type;
      list->list_Ref->next = NULL;
   }
   else
   {
      while ( p_Mess->next != NULL)
      {
         p_Mess = p_Mess->next;
      }

      p_List = (Reference*) malloc(sizeof(Reference));
      p_List->addr = ref.addr;
      p_List->type = ref.type;
      p_List->next = NULL;

      p_Mess->next = p_List;

   }

   return(Z_OK);
}

