/****************************************
 *      the Beos Reversing Engine       * 
 *              by Zadig                *
 *                                      *
 * This file miscellanous functions.    *
 ****************************************/

/*these files must always be includes before all others*/
#include <types.h>
#include "globals.h"
#include "structs.h"
#include "reveng_errors.h"
/* end these files must always be includes before all others*/

#include <zem_errors.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "misc.h" 

/***********************************************************
 *  Function: RE_GetHeader                 File: misc.c    *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Copies the header of reveng in the buffer sz_Buffer.    *
 ***********************************************************/ 
int32 RE_GetHeader(uint8 *sz_Buffer)
{
   sprintf((char*)sz_Buffer, 
           "**************************************\n"
           "*           RevEng  V%s           *\n"
           "*   The Reversing Engine by Zadig    *\n"
           "*     Compiled for %s   *\n"
           "**************************************\n\n", 
           RE_VERSION, HEADER_TARGET);  
   return(Z_OK);
}

/***********************************************************
 *  Function: RE_ReadString                File: misc.c    *
 ***********************************************************
 * Usage:                                                  *
 *                                                         *
 * Reads a string in the file and stops if there is a '<'  *
 * in it.                                                  *
 ***********************************************************/ 
int32 RE_ReadString(uint8 *sz_DestString, uint32 i_Size, FILE *f_file)
{
   int c_CurChar;
   uint16 i_Index;

   i_Index = 0;
   c_CurChar = fgetc(f_file);
   while( (c_CurChar != EOF) && ((uint8)c_CurChar != '\n')
         && ((uint8)c_CurChar != '<') 
         && (i_Index < (i_Size-1)) )
   {
      sz_DestString[i_Index] = c_CurChar;
      i_Index++;
      c_CurChar = fgetc(f_file);   
   }
   sz_DestString[i_Index] = '\0';

   if(i_Index == (i_Size-1))
   {
      return(FILE_ERROR_OVERFLOW);
   }
   else if((uint8)c_CurChar == '<')
   {
      c_CurChar = fgetc(f_file);
      while( ((uint8)c_CurChar != '\n') && (c_CurChar != EOF) )
      {
         c_CurChar = fgetc(f_file);
      }
   }

   if(c_CurChar == EOF)
   {
      return(FILE_ERROR_EOF);
   }

   return(Z_OK);
}

/************************************************************
 *  Function: RE_InitRegisters             File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 * Initialiszes all registers.                              *
 * "registers" must be ab array of REG_MAX Register struct. *
 ************************************************************/ 
int32 RE_InitRegisters(Register *registers, uint8 b_ResetAll)
{
   strcpy((char*)(registers[REG_EAX].name),"eax");
   registers[REG_EAX].isValid = Z_FALSE;
   registers[REG_EAX].type = RESULT_TYPE_INVALID;

   strcpy((char*)(registers[REG_ECX].name),"ecx");
   registers[REG_ECX].isValid = Z_FALSE;
   registers[REG_ECX].type = RESULT_TYPE_INVALID;

   strcpy((char*)(registers[REG_EDX].name),"edx");
   registers[REG_EDX].isValid = Z_FALSE;
   registers[REG_EDX].type = RESULT_TYPE_INVALID;

   strcpy((char*)(registers[REG_ESP].name),"esp");
   registers[REG_ESP].value = 0;
   registers[REG_ESP].isValid = Z_TRUE;
   registers[REG_ESP].type = RESULT_TYPE_DIRECT_VALUE;

   strcpy((char*)(registers[REG_EDI].name),"edi");
   registers[REG_EDI].isValid = Z_FALSE;
   registers[REG_EDI].type = RESULT_TYPE_INVALID;

   strcpy((char*)(registers[REG_ESI].name),"esi");
   registers[REG_ESI].isValid = Z_FALSE;
   registers[REG_ESI].type = RESULT_TYPE_INVALID;

   if(b_ResetAll == Z_TRUE)  
   {
      strcpy((char*)(registers[REG_EBX].name),"ebx");
      registers[REG_EBX].isValid = Z_FALSE;
      registers[REG_EBX].type = RESULT_TYPE_INVALID;

      strcpy((char*)(registers[REG_EBP].name),"ebp");
      registers[REG_EBP].isValid = Z_FALSE;
      registers[REG_EBP].type = RESULT_TYPE_INVALID;    
   }
   return(Z_OK);
}

/************************************************************
 *  Function: RE_SetNextCntx               File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 ************************************************************/ 
int32 RE_SetNextCntx(RegisterContext *context)
{
   int16 i;

   context->nextCntx = RE_DISABLED;
   for(i=0; i<MAX_REG_CNTX; i++)
   {
      if(context->context[i].free == Z_FALSE)
      {
         if( (context->nextCntx == RE_DISABLED) ||
               ( (context->nextCntx != RE_DISABLED) &&
                 (context->context[i].addr < context->context[context->nextCntx].addr) ) )
         {
            context->nextCntx = i;
         }
      }
   }     

   return(Z_OK); 
}

/************************************************************
 *  Function: RE_AddCntx                   File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 ************************************************************/ 
int32 RE_AddCntx(RegisterContext *context, Register *registers, uint32 i_Addr)
{
   uint16 i;
   uint16 i_CntxIndex;

   /* first we check a context does not exists with the same addr */
   for(i=0; i<MAX_REG_CNTX; i++)
   {
      if( (context->context[i].free == Z_FALSE) &&
          (context->context[i].addr == i_Addr) )
      {
         return(Z_OK); /* maybe we should return an error ?? */
      }
   }

   i_CntxIndex = MAX_REG_CNTX;
   for(i=0; i<MAX_REG_CNTX; i++)
   {
      if(context->context[i].free == Z_TRUE)
      {  
         i_CntxIndex = i;
         break;
      }
   }

   if(i_CntxIndex == MAX_REG_CNTX)
   {
      DBG_PRINT("Warning: Registers context array is full!\n");
   }
   else
   {
      context->context[i_CntxIndex].free = Z_FALSE;
      context->context[i_CntxIndex].addr = i_Addr;
      memcpy(&(context->context[i_CntxIndex].registers), registers, 
             REG_MAX * sizeof(Register));

      RE_SetNextCntx(context);  
   }

   return(Z_OK);
}

/************************************************************
 *  Function: RE_FreeCntx                  File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 ************************************************************/ 
int32 RE_FreeCntx(RegisterContext *context, uint32 i_CntxIndex)
{
   context->context[i_CntxIndex].free = Z_TRUE;
   return(Z_OK);
}

/************************************************************
 *  Function: RE_FreeCntx                  File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 ************************************************************/ 
int32 RE_EmptyCntx(RegisterContext *context)
{
   uint32 i;

   context->nextCntx = RE_DISABLED;
   for(i=0; i<MAX_REG_CNTX; i++)
      context->context[i].free = Z_TRUE;

   return(Z_OK);
}

/************************************************************
 *  Function: RE_GetParameters             File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 * Get parameters offsets and labels of the function.       *
 ************************************************************/ 
int32 RE_GetParameters(uint8 *sz_Function, 
                       Function_Params *parameters, 
                       LIST_Mapping *ParamsList)
{
   uint8    *psz_Name;
   uint8    sz_CurName[STRING_MAX_SIZE];
   uint16   i_Index;
   uint8    i_CurOffset;
   uint8    b_Found;
   Mapping  *p_CurParam;

   i_CurOffset = 0x04;
   parameters->nbParams = 0;
   parameters->isValid = Z_FALSE;
   psz_Name = sz_Function;

   if(strstr((char*)(psz_Name),"::") != NULL)
   {
      /* function is a C++ method, the class handle is the first param */
      parameters->functionIsC = Z_FALSE;
      parameters->isValid = Z_TRUE;
      parameters->param[0].offset = i_CurOffset;
      strcpy((char*)(parameters->param[0].name), "Object handle");
      i_CurOffset += 4;
      parameters->nbParams++;
   }
   else
      /* function is C code */
      parameters->functionIsC = Z_TRUE;

   while( (*psz_Name != '(') && (*psz_Name != '\0') )
      psz_Name++;

   if(*psz_Name == '\0')
      return(Z_OK);

   /* check if function deals with messages */  
   if(strstr((char*)psz_Name,"BMessage") != NULL)
      parameters->functionHasMessage = Z_TRUE;
   else
      parameters->functionHasMessage = Z_FALSE;

   psz_Name++;    
   if(strncmp((char*)psz_Name, "void)", 5) == 0)
      return(Z_OK);

   /* read all params of the function */
   while( (*psz_Name != '\0') &&
          (*psz_Name != ')') &&
          (parameters->nbParams < MAX_PARAMS) )
   {
      /* get next param name */
      i_Index = 0;
      while( (*psz_Name != ',') && 
             (*psz_Name != ')') &&
             (*psz_Name != '\0') )
      {
         sz_CurName[i_Index] = *psz_Name;
         psz_Name++;
         i_Index++;
      }
      sz_CurName[i_Index] = '\0';

      if(*psz_Name == ' ')
         psz_Name++;

      /* analyse the name */
      if(*psz_Name != '\0')
      {
         if( (strstr((char*)sz_CurName,"*") != NULL) ||
             (strstr((char*)sz_CurName,"&") != NULL) ||
             (strstr((char*)sz_CurName,"char") != NULL) ||
             (strstr((char*)sz_CurName,"int") != NULL) ||
             (strstr((char*)sz_CurName,"long") != NULL) )
         {
            parameters->isValid = Z_TRUE;
            parameters->param[parameters->nbParams].offset = i_CurOffset;
            strcpy((char*)(parameters->param[parameters->nbParams].name), (char*)sz_CurName);
            i_CurOffset += 4;
            parameters->nbParams++;
         }         
         else /* look in params list */
         {
            p_CurParam = ParamsList->list_Map;
            b_Found = Z_FALSE;
            while( (p_CurParam != NULL) &&
                   (b_Found == Z_FALSE) )
            {
               if(strstr((char*)sz_CurName, (char*)p_CurParam->label) != NULL)
                   /* (sz_CurName[strlen(p_CurParam->label) - 1] < 'A') ) */
               {
                  b_Found = Z_TRUE;
                  
                  parameters->isValid = Z_TRUE;
                  parameters->param[parameters->nbParams].offset = p_CurParam->addr;
                  strcpy((char*)(parameters->param[parameters->nbParams].name), 
                         (char*)p_CurParam->label);
                  i_CurOffset += p_CurParam->addr;
                  parameters->nbParams++;
               }
               p_CurParam = p_CurParam->next;
            }
            
            if(b_Found == Z_FALSE)
               break;
         }
         psz_Name++;
      }
   }
   
   if(parameters->nbParams >= MAX_PARAMS)
   {
      DBG_PRINT("RE_GetParameters Warning: Too much params!\n");
      DBG_PRINT(sz_Function);
      DBG_PRINT("\n");
   }
   return(Z_OK);
}

/************************************************************
 *  Function: RE_GetParamIndex             File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 * Get parameters offsets and labels of the function.       *
 ************************************************************/ 
uint8 RE_GetParamIndex(Function_Params *parameters, uint16 i_Offset)
{  
   uint16 i;
   uint8 i_ret;

   i_ret = 0;  
   if( (parameters->isValid == Z_FALSE) ||
       (parameters->nbParams >= MAX_PARAMS) )
   {
      return(i_ret);
   }

   i=0;
   while(i < MAX_PARAMS)
   {
      if(parameters->param[i].offset == i_Offset)
      {
         i_ret = i+1;
         break;
      }
      i++;
   }

   if(i>= parameters->nbParams)
      i_ret = 0;

   return(i_ret);
}

/************************************************************
 *  Function: RE_ParamIsMessage            File: misc.c     *
 ************************************************************
 * Usage:                                                   *
 *                                                          *
 * Tells if a parameter is a message or not.                *
 ************************************************************/ 
uint8 RE_ParamIsMessage(Function_Params *parameters, uint8 i_Index)
{
   if( (i_Index < 1) || (parameters->isValid == Z_FALSE) )
      return(Z_FALSE);

   if(strstr((char*)(parameters->param[i_Index-1].name), "BMessage") != NULL)
      return(Z_TRUE);
   else
      return(Z_FALSE);
}
